"""Useful functions/class/types for defining loss functions, metrics, model loading and data augmentation"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/01_model_helpers.ipynb.

# %% auto 0
__all__ = ['BudgetType', 'load_module', 'load_yaml', 'AbstractModel']

# %% ../../nbs/utils/01_model_helpers.ipynb 3
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import xarray as xr
import yaml

import importlib.util as import_utils
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Union, List, Dict

# %% ../../nbs/utils/01_model_helpers.ipynb 5
def load_module(
  module_name: str, # The name of the module to load
  module_path: Path, # The path to the module
) -> object: # The loaded module
    "Load a module from a file path."
    spec = import_utils.spec_from_file_location(module_name, module_path)
    module = import_utils.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module

# %% ../../nbs/utils/01_model_helpers.ipynb 6
def load_yaml(
  file_path: Path, # The path to the YAML file
) -> Dict[str, Union[list[str], str]]: # The loaded module
    "Load a yaml file."
    with open(file_path, 'r') as file:
        return yaml.safe_load(file)

# %% ../../nbs/utils/01_model_helpers.ipynb 8
class AbstractModel(ABC):
  """An abstract class for models"""
  @abstractmethod
  def predict(self, x: xr.Dataset) -> xr.Dataset:
    pass
  
  @abstractmethod
  def contributions(self, x: xr.Dataset) -> xr.Dataset:
    pass
  
  def __repr__(self) -> str:
    return f"Model"

# %% ../../nbs/utils/01_model_helpers.ipynb 11
BudgetType = Union[Dict[str, float], xr.Dataset] # type alias for budget data



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/budget_optimizer',
                'doc_host': 'https://optimizer.mattreda.pro',
                'git_url': 'https://github.com/redam94/budget_optimizer',
                'lib_path': 'budget_optimizer'},
  'syms': { 'budget_optimizer.optimizer': { 'budget_optimizer.optimizer.BaseOptimizer': ( 'optimizer.html#baseoptimizer',
                                                                                          'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.BaseOptimizer.__init__': ( 'optimizer.html#baseoptimizer.__init__',
                                                                                                   'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.BaseOptimizer._load_config': ( 'optimizer.html#baseoptimizer._load_config',
                                                                                                       'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.BaseOptimizer._load_loss_fn': ( 'optimizer.html#baseoptimizer._load_loss_fn',
                                                                                                        'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.BaseOptimizer._load_optimizer_array_to_budget': ( 'optimizer.html#baseoptimizer._load_optimizer_array_to_budget',
                                                                                                                          'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.BaseOptimizer._optimizer_fn': ( 'optimizer.html#baseoptimizer._optimizer_fn',
                                                                                                        'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.BaseOptimizer.optimize': ( 'optimizer.html#baseoptimizer.optimize',
                                                                                                   'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.BaseOptimizer.reload_config': ( 'optimizer.html#baseoptimizer.reload_config',
                                                                                                        'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.OptunaBudgetOptimizer': ( 'optimizer.html#optunabudgetoptimizer',
                                                                                                  'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.OptunaBudgetOptimizer.__init__': ( 'optimizer.html#optunabudgetoptimizer.__init__',
                                                                                                           'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.OptunaBudgetOptimizer._opt_fn': ( 'optimizer.html#optunabudgetoptimizer._opt_fn',
                                                                                                          'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.OptunaBudgetOptimizer.optimize': ( 'optimizer.html#optunabudgetoptimizer.optimize',
                                                                                                           'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.ScipyBudgetOptimizer': ( 'optimizer.html#scipybudgetoptimizer',
                                                                                                 'budget_optimizer/optimizer.py'),
                                            'budget_optimizer.optimizer.ScipyBudgetOptimizer.optimize': ( 'optimizer.html#scipybudgetoptimizer.optimize',
                                                                                                          'budget_optimizer/optimizer.py')},
            'budget_optimizer.utils.model_classes': { 'budget_optimizer.utils.model_classes.BaseBudgetModel': ( 'utils/model_classes.html#basebudgetmodel',
                                                                                                                'budget_optimizer/utils/model_classes.py'),
                                                      'budget_optimizer.utils.model_classes.BaseBudgetModel.__init__': ( 'utils/model_classes.html#basebudgetmodel.__init__',
                                                                                                                         'budget_optimizer/utils/model_classes.py'),
                                                      'budget_optimizer.utils.model_classes.BaseBudgetModel._get_budget_to_data': ( 'utils/model_classes.html#basebudgetmodel._get_budget_to_data',
                                                                                                                                    'budget_optimizer/utils/model_classes.py'),
                                                      'budget_optimizer.utils.model_classes.BaseBudgetModel._get_model_loader': ( 'utils/model_classes.html#basebudgetmodel._get_model_loader',
                                                                                                                                  'budget_optimizer/utils/model_classes.py'),
                                                      'budget_optimizer.utils.model_classes.BaseBudgetModel.contributions': ( 'utils/model_classes.html#basebudgetmodel.contributions',
                                                                                                                              'budget_optimizer/utils/model_classes.py'),
                                                      'budget_optimizer.utils.model_classes.BaseBudgetModel.predict': ( 'utils/model_classes.html#basebudgetmodel.predict',
                                                                                                                        'budget_optimizer/utils/model_classes.py')},
            'budget_optimizer.utils.model_helpers': { 'budget_optimizer.utils.model_helpers.AbstractModel': ( 'utils/model_helpers.html#abstractmodel',
                                                                                                              'budget_optimizer/utils/model_helpers.py'),
                                                      'budget_optimizer.utils.model_helpers.AbstractModel.__repr__': ( 'utils/model_helpers.html#abstractmodel.__repr__',
                                                                                                                       'budget_optimizer/utils/model_helpers.py'),
                                                      'budget_optimizer.utils.model_helpers.AbstractModel.contributions': ( 'utils/model_helpers.html#abstractmodel.contributions',
                                                                                                                            'budget_optimizer/utils/model_helpers.py'),
                                                      'budget_optimizer.utils.model_helpers.AbstractModel.predict': ( 'utils/model_helpers.html#abstractmodel.predict',
                                                                                                                      'budget_optimizer/utils/model_helpers.py'),
                                                      'budget_optimizer.utils.model_helpers.load_module': ( 'utils/model_helpers.html#load_module',
                                                                                                            'budget_optimizer/utils/model_helpers.py'),
                                                      'budget_optimizer.utils.model_helpers.load_yaml': ( 'utils/model_helpers.html#load_yaml',
                                                                                                          'budget_optimizer/utils/model_helpers.py')},
            'budget_optimizer.utils.search_space_helper': { 'budget_optimizer.utils.search_space_helper.ConstrainedSearchSpace': ( 'utils/search_space_helpers.html#constrainedsearchspace',
                                                                                                                                   'budget_optimizer/utils/search_space_helper.py'),
                                                            'budget_optimizer.utils.search_space_helper.ConstrainedSearchSpace.__call__': ( 'utils/search_space_helpers.html#constrainedsearchspace.__call__',
                                                                                                                                            'budget_optimizer/utils/search_space_helper.py'),
                                                            'budget_optimizer.utils.search_space_helper.ConstrainedSearchSpace.__post_init__': ( 'utils/search_space_helpers.html#constrainedsearchspace.__post_init__',
                                                                                                                                                 'budget_optimizer/utils/search_space_helper.py'),
                                                            'budget_optimizer.utils.search_space_helper.Trial': ( 'utils/search_space_helpers.html#trial',
                                                                                                                  'budget_optimizer/utils/search_space_helper.py'),
                                                            'budget_optimizer.utils.search_space_helper.Trial.suggest_float': ( 'utils/search_space_helpers.html#trial.suggest_float',
                                                                                                                                'budget_optimizer/utils/search_space_helper.py')}}}



"""Generate constrained search space. Alert if constraints are not satisfiable."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/02_search_space_helpers.ipynb.

# %% auto 0
__all__ = ['Trial', 'ConstrainedSearchSpace']

# %% ../../nbs/utils/02_search_space_helpers.ipynb 3
import numpy as np
from typing import Generator, List, Tuple, Protocol
from dataclasses import dataclass
from collections import OrderedDict

# %% ../../nbs/utils/02_search_space_helpers.ipynb 4
class Trial(Protocol):
    "Protocol for a trial object"
    @staticmethod
    def suggest_float(name: str, *args, **kwargs) -> float:
        ...

# %% ../../nbs/utils/02_search_space_helpers.ipynb 5
@dataclass
class ConstrainedSearchSpace:
    """
    A class that generates a search space with constraints
    """
    bounds: dict[str, tuple[float, float]] # bounds of each channel
    constraint: tuple[float, float] # constraint of the sum of all channels

    def __post_init__(self):
        self.bounds = OrderedDict(sorted(self.bounds.items(), key=lambda x: x[1]))
    
    def __call__(
        self, 
        trial: Trial # trial object
        ) -> dict[str, float]: # selected budget
        "Sample from constrained search space"
        selected_budget = {}
        bounds_values = list(self.bounds.values())
        bounds_items = list(self.bounds.items())
        for n, (name, bound) in enumerate(bounds_items[:-1]):
            curr_total = sum(selected_budget.values())
            new_min_bound = self.constraint[0]-(curr_total+sum(b[1] for b in bounds_values[n+1:]))
            new_max_bound = self.constraint[1]-(curr_total+sum(b[0] for b in bounds_values[n+1:]))
            updated_bounds = (
                    max(bound[0], new_min_bound), 
                    min(bound[1], new_max_bound)
            )
            selection = trial.suggest_float(name, *updated_bounds)
            selected_budget[name] = selection
        last = bounds_values[-1]
        last_name = bounds_items[-1][0]
        choice = trial.suggest_float(
            last_name, 
            max(last[0], self.constraint[0]-sum(selected_budget.values())),
            min(last[1], self.constraint[1]-sum(selected_budget.values()))
        )
        selected_budget[last_name] = choice
        return selected_budget



"""Model classes"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/00_model_classes.ipynb.

# %% auto 0
__all__ = ['BaseBudgetModel']

# %% ../../nbs/utils/00_model_classes.ipynb 5
from abc import ABC, abstractmethod
from pathlib import Path
from collections.abc import Sequence
from typing import (
  Callable, Generic, 
  TypeVar, Union, 
  Protocol, Dict,
  TypeAlias)
import types

import xarray as xr

from budget_optimizer.utils.model_helpers import (
  load_module,
  AbstractModel,
  BudgetType
)

# %% ../../nbs/utils/00_model_classes.ipynb 6
class BaseBudgetModel(AbstractModel):
    """
    Abstract class for all models
    """
    _FUNCTION_MODULE_NAME = "model_config.py"
    
    def __init__(
      self, 
      model_name: str, # Name used to identify the model
      model_kpi: str, # Key performance indicator output by the model predict
      model_path: str|Path, # Path to the model artifact
    ):
        self.model_name: str = model_name
        self.model_kpi: str = model_kpi
        self.model_path: Path = model_path if isinstance(model_path, Path) else Path(model_path)
        self._model: AbstractModel = self._get_model_loader()(model_path)
        self._budget_to_data: xr.Dataset = self._get_budget_to_data()

    def _get_model_loader(self) -> Callable[str|Path, AbstractModel]:
        """
        Get the function to load the model from the path
        """
        module = load_module(
            self._FUNCTION_MODULE_NAME.replace(".py", ""),
            self.model_path / self._FUNCTION_MODULE_NAME
            )
        return module.model_loader
    
    def _get_budget_to_data(self) -> Callable[BudgetType, xr.Dataset]:
        """
        Get the mapping from budget keys to data keys
        """
        
        module = load_module(
            self._FUNCTION_MODULE_NAME.replace(".py", ""), 
            self.model_path / self._FUNCTION_MODULE_NAME
            )
        
        return module.budget_to_data
    
    def predict(
        self, 
        budget: BudgetType # Budget
        ) -> xr.DataArray: # Predicted target variable
        """
        Predict the target variable from the input data
        """
        data = self._budget_to_data(budget, self._model)
        return self._model.predict(data)
    
    
    def contributions(
        self, 
        budget: BudgetType # Budget
        ) -> xr.Dataset: # Contributions of the input data to the target variable
        """
        Get the contributions of the input data to the target variable
        """
        data = self._budget_to_data(budget, self._model)
        return self._model.contributions(data)



"""Optimizers for models"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_optimizer.ipynb.

# %% auto 0
__all__ = ['BaseOptimizer', 'ScipyBudgetOptimizer', 'OptunaBudgetOptimizer']

# %% ../nbs/00_optimizer.ipynb 5
import numpy as np
import xarray as xr
import pandas as pd
import scipy.optimize as opt
import optuna

from .utils.model_classes import BaseBudgetModel
from budget_optimizer.utils.model_helpers import (
  load_module,
  load_yaml,
  BudgetType, 
  AbstractModel
)
from .utils.search_space_helper import ConstrainedSearchSpace
from pathlib import Path
from abc import ABC, abstractmethod

# %% ../nbs/00_optimizer.ipynb 6
class BaseOptimizer(ABC):
    """Optimizer wrapper for the pyswarms package"""
    _CONFIG_YAML = 'optimizer_config.yaml'
    _MODULE_FILE = "optimizer_config.py"
    
    def __init__(
        self, 
        model: BaseBudgetModel, # The model to optimize
        config_path: str|Path # Path to the configuration files
        ):
        
        self.model: BaseBudgetModel = model
        self._config_path: Path = Path(config_path) if isinstance(config_path, str) else config_path
        self.optimal_budget: BudgetType = None
        self.optimal_prediction: xr.DataArray = None
        self.optimal_contribution: xr.Dataset = None
        self.sol = None
        self._config = self._load_config()
        self._loss_fn = self._load_loss_fn()
        self._optimizer_array_to_budget = self._load_optimizer_array_to_budget()
        
    def _load_config(self):
        config = load_yaml(self._config_path / self._CONFIG_YAML)
        return config
    
    def reload_config(self):
        self._config = self._load_config()
        return self
    
    def _load_loss_fn(self):
        """Load the loss function from the config file"""
        module = load_module(self._MODULE_FILE.replace(".py", ""), self._config_path / self._MODULE_FILE)
        return module.loss_fn
    
    def _load_optimizer_array_to_budget(self):
        """Convert the optimizer array to a budget"""
        module = load_module(self._MODULE_FILE.replace(".py", ""), self._config_path / self._MODULE_FILE)
        return module.optimizer_array_to_budget
    
    def _optimizer_fn(self, x: np.ndarray):
        """Optimizer step"""
        budget = self._optimizer_array_to_budget(x)
        prediction = self.model.predict(budget)
        loss = self._loss_fn(prediction, **self._config['loss_fn_kwargs'])
        return loss
    
    @abstractmethod
    def optimize(
        self, 
        bounds: list[tuple[float, float]], # Bounds for the optimizer
        constraints: None|opt.LinearConstraint|tuple[float, float] = None, # Constraints for the optimizer
        **kwargs: dict # Additional arguments for the optimizer
        ):
        """Optimize the model"""
        raise NotImplementedError("This method should be implemented in the child class")

# %% ../nbs/00_optimizer.ipynb 7
class ScipyBudgetOptimizer(BaseOptimizer):
    """Optimizer wrapper for the pyswarms package"""
    
    def optimize(
        self, 
        bounds: list[tuple[float, float]], # Bounds for the optimizer
        constraints: None|opt.LinearConstraint, # Constraints for the optimizer
        init_pos: np.ndarray, # Initial position of the optimizer
        ):
        """Optimize the model"""
        import warnings
        warnings.filterwarnings("ignore")
        self.sol = opt.minimize(
            self._optimizer_fn, init_pos,
            method='trust-constr', 
            bounds=bounds, 
            constraints=constraints
            )
        if not self.sol.success:
            raise Exception(f"Optimization failed: {self.sol.message}")
        
        self.optimal_budget = self._optimizer_array_to_budget(self.sol.x)
        self.optimal_prediction = self.model.predict(self.optimal_budget) # The optimizer minimizes the cost, so we need to negate it
        self.optimal_contribution = self.model.contributions(self.optimal_budget)
        return self

# %% ../nbs/00_optimizer.ipynb 15
from typing import Literal

# %% ../nbs/00_optimizer.ipynb 16
class OptunaBudgetOptimizer(BaseOptimizer):
    def __init__(
        self, 
        model: BaseBudgetModel, # The model to optimize
        config_path: str|Path, # Path to the configuration files
        objective_name: str = "loss", # Name of the objective
        direction: Literal["maximize", "minimize"] = "maximize", # Direction of the optimization
        sampler: optuna.samplers.BaseSampler = optuna.samplers.TPESampler, # Sampler for the optimization
        pruner: optuna.pruners.BasePruner|None = None, # Pruner for the optimization
        tol: float = 1e-3, # Tolerance for the constraints
        percent_out_tolerance: float = 0.1, # Percentage of the budget trials that can be outside the constraints
        sampler_kwargs: dict|None = None, # Additional arguments for the sampler
        pruner_kwargs: dict|None = None, # Additional arguments for the pruner
        ):
        super().__init__(model, config_path)
        self.objective_name = objective_name
        self.study = None
        self._direction = direction
        self.__tol = tol
        self.__percent_out_tolerance = percent_out_tolerance
        self.__sampler = sampler(**sampler_kwargs)
        self.__pruner = pruner(**pruner_kwargs) if not pruner is None else None
        
        
    #def _constraints(self, trial):
    #    return trial.user_attrs["constraint"]
    
    
    def _opt_fn(self, trial):
        budget = self.search_space(trial)
        #budget = self._optimizer_array_to_budget([budget[name] for name in ])
        total_budget = sum(v for v in budget.values())
        
        trial.set_user_attr("budget", budget)
        trial.set_user_attr("total_budget", total_budget)
        prediction = self.model.predict(budget)
        
        
        loss = -self._loss_fn(prediction, **self._config['loss_fn_kwargs'])
        return loss
        
        

    def optimize(
        self, 
        bounds: dict[str, tuple[float, float]], # Bounds for the optimizer
        constraints: None|tuple = None, # Constraints for the optimizer
        timeout: int = 60,
        n_trials: int = 100,
        storage: str|None = "sqlite:///db.sqlite3",
        study_name: str = "optimizer",
        n_jobs: int = 1
    ):
        """Optimize the model"""
            
        self.study = optuna.create_study(
            storage=storage,  # Specify the storage URL here.
            study_name=study_name,
            direction=self._direction,
            sampler=self.__sampler,
            pruner=self.__pruner)
        self.study.set_metric_names([self.objective_name])
        if constraints is None:
            constraints = (-np.inf, np.inf)
        self.search_space = ConstrainedSearchSpace(bounds, constraints)
        self.study.optimize(
            self._opt_fn, 
            n_trials=n_trials, 
            timeout=timeout,
            n_jobs=n_jobs)
        
        self.sol = self.study.best_trial
        self.optimal_budget = self.sol.params
        self.optimal_prediction = self.model.predict(self.optimal_budget)
        
